@@ -1,4 +1,4 @@
-from flask import Flask, jsonify, request, Response, abort
+from flask import Flask, jsonify, request, Response, abort, Blueprint
 from flask_socketio import SocketIO
 from flask_cors import CORS
 
@@ -31,23 +31,26 @@ from ..stream import Stream
 from ..source import Source
 
 
-app = Flask("hw-server")
+app = Flask(__name__)
 CORS(app)
 sio = SocketIO(app)
 
+api_version = '/v0'
 
-@app.get("/")
+api = Blueprint('api', __name__)
+
+@api.get("/")
 def index():
     return jsonify(msg=f"says hello and {random.random()}")
 
 
-@app.get("/ping")
+@api.get("/ping")
 def ping():
 
     return jsonify(pong=random.randint(1, 100))
 
 
-@app.get("/start")
+@api.get("/start")
 def start():
 
     """route to start the stream provided a stream_name in the url params
@@ -59,12 +62,14 @@ def start():
     if request.args:
         # NB for request arg this is an ImmutableMultiDict from werkzeug, so can access using ['key'] format
         # but it generates it's own error messages if a key is not found, this is useful in general, but complicates
-        # things in this case, so we convert to regular dictionary, so this keeps code handling args 
+        # things in this case, so we convert to regular dictionary, so this keeps code handling args
         # keys similar to handling json keys
         data = dict(request.args)
     else:
         return (
-            jsonify(msg=f"RequestError: request must contain url arguments (probably missing 'stream_name')"),
+            jsonify(
+                msg=f"RequestError: request must contain url arguments (probably missing 'stream_name')"
+            ),
             400,
         )
 
@@ -90,7 +95,6 @@ def start():
             msg=f"TypeError: stream name (stream_name) must be an integer; supplied value was of type {type(stream_name)}"
         )
 
-
     for stream in streams:
         if stream.name == stream_name:
             try:
@@ -104,12 +108,14 @@ def start():
                 )
 
     return (
-        jsonify(msg=f"ValueError: seems like stream '{stream_name}' has not been created...?"),
+        jsonify(
+            msg=f"ValueError: seems like stream '{stream_name}' has not been created...?"
+        ),
         404,
     )
 
 
-@app.get("/stop")
+@api.get("/stop")
 def stop():
 
     """route to stop the stream provided a stream_name in the url params
@@ -121,12 +127,14 @@ def stop():
     if request.args:
         # NB for request arg this is an ImmutableMultiDict from werkzeug, so can access using ['key'] format
         # but it generates it's own error messages if a key is not found, this is useful in general, but complicates
-        # things in this case, so we convert to regular dictionary, so this keeps code handling args 
+        # things in this case, so we convert to regular dictionary, so this keeps code handling args
         # keys similar to handling json keys
         data = dict(request.args)
     else:
         return (
-            jsonify(msg=f"RequestError: request must contain url arguments (probably missing 'stream_name')"),
+            jsonify(
+                msg=f"RequestError: request must contain url arguments (probably missing 'stream_name')"
+            ),
             400,
         )
 
@@ -152,7 +160,6 @@ def stop():
             msg=f"TypeError: stream name (stream_name) must be an integer; supplied value was of type {type(stream_name)}"
         )
 
-
     for stream in streams:
         if stream.name == stream_name:
             try:
@@ -166,12 +173,14 @@ def stop():
                 )
 
     return (
-        jsonify(msg=f"ValueError: seems like stream '{stream_name}' has not been created...?"),
+        jsonify(
+            msg=f"ValueError: seems like stream '{stream_name}' has not been created...?"
+        ),
         404,
     )
 
 
-@app.get("/stream_status")
+@api.get("/stream_status")
 def stream_status():
     """a general purpose route to enable quick acquisition of a stream state
 
@@ -186,7 +195,9 @@ def stream_status():
         data = dict(request.args)
     else:
         return (
-            jsonify(msg=f"RequestError: request must contain url arguments (probably missing 'stream_name')"),
+            jsonify(
+                msg=f"RequestError: request must contain url arguments (probably missing 'stream_name')"
+            ),
             400,
         )
 
@@ -212,7 +223,6 @@ def stream_status():
             msg=f"TypeError: stream name (stream_name) must be an integer; supplied value was of type {type(stream_name)}"
         )
 
-
     for stream in streams:
         if stream.name == stream_name:
             try:
@@ -227,13 +237,81 @@ def stream_status():
                 )
 
     return (
-        jsonify(msg=f"ValueError: seems like stream '{stream_name}' has not been created...?"),
+        jsonify(
+            msg=f"ValueError: seems like stream '{stream_name}' has not been created...?"
+        ),
         404,
     )
 
 
+@api.get("/source")
+def get_source():
+    """a GET route to enable acquisition of a source instance state
 
-@app.patch("/freq")
+    uses the source_status @property (not a method to call)
+
+    returns the state of one or all Source instance(s) in the 'sources' list
+    so the client can retrieve latest state
+
+    """
+    # first, check the request has parameter arguments
+    if request.args:
+        # NB for request arg this is an ImmutableMultiDict from werkzeug, so can access using ['key'] format
+        # this keeps code handling args similar to handling json
+        data = dict(request.args)
+
+        try:
+            stream_name = data["stream_name"]
+            if not stream_name:
+                raise ValueError("'stream_name' must not be empty")
+
+        except KeyError as e:
+            return (
+                jsonify(msg=f"missing key: {str(e)}"),
+                400,
+            )
+        except ValueError as e:
+            return (
+                jsonify(msg=f"value error: {str(e)}"),
+                400,
+            )
+
+        # check that the value of stream_name is of type 'str'
+        if not isinstance(stream_name, str):
+            return jsonify(
+                msg=f"TypeError: stream name (stream_name) must be an integer; supplied value was of type {type(stream_name)}"
+            )
+
+        for source in sources:
+            if source.name == stream_name:
+                try:
+                    r = source.get_source_state
+                    return jsonify(r)
+
+                # handle an error which as of writing is of unknown type...
+                except Exception as e:
+                    return (
+                        jsonify(msg=f"{str(e)}"),
+                        400,
+                    )
+
+        return (
+            jsonify(
+                msg=f"ValueError: seems like stream '{stream_name}' has not been created...?"
+            ),
+            404,
+        )
+    else:
+        # NOTE this will be dropped to return all source instances in sources list eveutally
+        return (
+            jsonify(
+                msg=f"RequestError: request must contain url arguments (probably missing 'stream_name')"
+            ),
+            400,
+        )
+
+
+@api.patch("/freq")
 def freq():
     """PATCH that is sent the new required value for the freq of stream new_freq
     and use the stream.set_freq(new_freq) setter
@@ -302,7 +380,7 @@ def freq():
             )
 
 
-@app.patch("/start_burst")
+@api.patch("/start_burst")
 def start_burst():
     """trigger a burst by setting the stream.burst_mode to True via
     calling the stream.start_burst() setter method
@@ -366,7 +444,7 @@ def start_burst():
             )
 
 
-@app.patch("/burst_freq")
+@api.patch("/burst_freq")
 def burst_freq():
     """set the burst frequency by calling stream.set_burst_freq method"""
 
@@ -432,7 +510,7 @@ def burst_freq():
             )
 
 
-@app.patch("/burst_vol")
+@api.patch("/burst_vol")
 def burst_vol():
     """set the burst volume by calling stream.set_burst_vol method"""
 
@@ -498,28 +576,107 @@ def burst_vol():
             )
 
 
-@app.get("/error_on")
-def error_on():
-    stream = random.choice(streams)
-    stream.set_error_mode_on()
+@api.patch("/source")
+def patch_source():
 
-    return jsonify(msg=f"error mode set for stream {stream.name}")
+    # first, check the request has a json data payload
+    if request.json:
+        data = request.json
+    else:
+        return (
+            jsonify(msg=f"RequestError: request must contain a json payload"),
+            400,
+        )
 
+    # then, check all required keys are present and have a value
+    try:
+        source_name = data["stream_name"]
+        if not source_name:
+            raise ValueError("'stream_name' must not be empty")
 
-@app.post("/add_field")
-def add_word():
+        config_area = data["config_area"]
+        if not config_area:
+            raise ValueError(f"'config_area' cannot be blank for a patch")
 
-    data = request.json
+        setting = data["setting"]
+        if not setting:
+            raise ValueError("'setting' must not be empty")
 
-    this_domain = data["source"]
+        new_setting_val = data["new_setting_val"]
+
+    except KeyError as e:
+        return (
+            jsonify(msg=f"missing key: {str(e)}"),
+            400,
+        )
+    except ValueError as e:
+        return (
+            jsonify(msg=f"value error: {str(e)}"),
+            400,
+        )
+
+    except Exception as e:
+        return (
+            jsonify(msg=f"unkown error: {str(e)}"),
+            400,
+        )
+    # OPTIONAL KEYS
+    try:
+        field_name = data["field_name"]
+        if not field_name:
+            raise ValueError("'field_name' must not be empty")
+    except KeyError as e:
+        field_name = None
+    except ValueError as e:
+        # but if it is there it needs a value!
+        return (
+            jsonify(msg=f"value error: {str(e)}"),
+            400,
+        )
+
+    # check that the value of stream_name is of type 'str'
+    # NB Stream class instances handle type checking for their own properties and methods
+    if not isinstance(source_name, str):
+        return jsonify(
+            msg=f"TypeError: source name {source_name} must be an string; supplied value was of type {type(source_name)}"
+        )
 
-    r = "huh"
     for source in sources:
-        if this_domain == source.name:
-            r = source.set_field(data)
-            break
+        if source.name == source_name:
+            try:
+                # print("made it to set source call in server")
+                source.set_source_element(
+                    config_area=config_area,
+                    field_name=field_name,
+                    setting=setting,
+                    new_setting_val=new_setting_val,
+                )
 
-    return jsonify(msg=r)
+            except ValueError as e:
+                return (
+                    jsonify(msg=f"{str(e)}"),
+                    400,
+                )
+            except TypeError as e:
+                return (
+                    jsonify(msg=f"{str(e)}"),
+                    400,
+                )
+            except Exception as e:
+                return (
+                    jsonify(msg=f"unknown error {str(e)}"),
+                    400,
+                )
+            else:
+                r = source.get_source_state
+                return jsonify(r)
+        else:
+            return (
+                jsonify(
+                    msg=f"ValueError: seems like stream '{source_name}' has not been created...?"
+                ),
+                404,
+            )
 
 
 @app.route("/ui", defaults={"path": ""})
@@ -563,6 +720,7 @@ def connect_hndlr():
 streams = []
 sources = []
 
+app.register_blueprint(api, url_prefix=f"/api/{api_version}")
 
 def run(selected_sources):
     """ """
@@ -571,7 +729,11 @@ def run(selected_sources):
         try:
             source = Source(selected_source)
         except FileNotFoundError:
-            print(Fore.YELLOW + f"   source name '{selected_source}' not resolved in schema lookup" + Style.RESET_ALL)
+            print(
+                Fore.YELLOW
+                + f"   source name '{selected_source}' not resolved in schema lookup"
+                + Style.RESET_ALL
+            )
             print()
             continue
         sources.append(source)
@@ -589,14 +751,14 @@ def run(selected_sources):
     print(
         Fore.GREEN
         + Style.BRIGHT
-        + f"STREAMS: http://127.0.0.1:{port}"
+        + f"STREAMS: http://127.0.0.1:{port}/api/v0"
         + Style.RESET_ALL
     )
     print(
         Fore.CYAN + Style.BRIGHT + f"UI: http://127.0.0.1:{port}/ui" + Style.RESET_ALL
     )
     print()
-    print( Style.DIM + "(CTRL-C to stop)" + Style.RESET_ALL)
+    print(Style.DIM + "(CTRL-C to stop)" + Style.RESET_ALL)
 
     sio.run(app, debug=False, port=port)
 
@@ -608,7 +770,10 @@ def run(selected_sources):
             stream.stop()
             print(Fore.RED + f"   stopped stream '{stream.name}'" + Style.RESET_ALL)
         except ValueError as e:
-            print(Fore.RED + f"   stream '{stream.name}' already stopped" + Style.RESET_ALL)
-    
-    print()
+            print(
+                Fore.RED
+                + f"   stream '{stream.name}' already stopped"
+                + Style.RESET_ALL
+            )
 
+    print()
